#define _POSIX_C_SOURCE     200809L

#include  <stdio.h>
#include  <stdbool.h>
#include  <stdlib.h>
#include  <sys/types.h>
#include  <sys/stat.h>
#include  <dirent.h>
#include  <sys/param.h>
#include  <time.h>
#include  <string.h>

#include  <getopt.h>

#define	OPTLIST		"r"

//  AS DISCUSSED IN WORKSHOP-7 :
#define	CHECK_ALLOC(p) if(p == NULL) { perror(__func__); exit(EXIT_FAILURE); }

typedef struct {
    char        *pathname;
    time_t      mtime;
} FILES;

FILES           *files  = NULL;
int             nfiles  = 0;

void list_all_files(void)
{
    for(int n=0 ; n<nfiles ; ++n) {
        char *timestr = ctime(&files[n].mtime);

        timestr[24] = '\0';		// remove pesky trailing newline
        printf( "%s\t%s\n", timestr, files[n].pathname );
    }
}

void scan_directory(char *dirname)        
{
    DIR             *dirp;
    struct dirent   *dp;

//  ENSURE THAT WE CAN OPEN (read-only) THE REQUIRED DIRECTORY
    dirp       = opendir(dirname);
    if(dirp == NULL) {
        perror( dirname );
        exit(EXIT_FAILURE);
    }

//  READ FROM THE REQUIRED DIRECTORY, UNTIL WE REACH ITS END
    while((dp = readdir(dirp)) != NULL) {  
        struct stat     stat_info;
        char            pathname[MAXPATHLEN];

        sprintf(pathname, "%s/%s", dirname, dp->d_name);

//  DETERMINE ATTRIBUTES OF THIS DIRECTORY ENTRY
        if(stat(pathname, &stat_info) != 0) {
            perror( pathname );
            exit(EXIT_FAILURE);
        }

//  EXTEND OUR ARRAY OF STRUCTURES BY ONE ELEMENT
        files                   = realloc(files, (nfiles+1)*sizeof(files[0]));
	CHECK_ALLOC(files);			// ensure allocation was OK

//  REMEMBER (COPY) THIS ELEMENT'S RELATIVE PATHNAME
        files[nfiles].pathname  = strdup(pathname);
	CHECK_ALLOC(files[nfiles].pathname);	// ensure allocation was OK

//  REMEMBER THIS ELEMENT'S MODIFICATION TIME
        files[nfiles].mtime     = stat_info.st_mtime;
        ++nfiles;
    }

//  CLOSE THE DIRECTORY
    closedir(dirp);
}

//  SORT STRUCTURES IN ASCENDING ORDER
int compare_mtime_ascending(const void *v1, const void *v2)
{
   FILES *f1 = (FILES *)v1;
   FILES *f2 = (FILES *)v2;

    return(f1->mtime - f2->mtime);
}

//  SORT STRUCTURES IN DECENDING ORDER
int compare_mtime_descending(const void *v1, const void *v2)
{
   FILES *f1 = (FILES *)v1;
   FILES *f2 = (FILES *)v2;

    return(f2->mtime - f1->mtime);
}

int main(int argc, char *argv[])
{
    bool	reverse	= false;
    int		opt;

//  PROCESS COMMAND-LINE OPTIONS  (see Lecture 18)
    opterr	= 0;
    while((opt = getopt(argc, argv, OPTLIST)) != -1) {
	if(opt == 'r') {
            reverse = !reverse;
	    ++argv;
	    --argc;
	}
        else {
            argc = -1;
	}
    }

//  WAS AN INVALID COMMAND-LINE OPTION PROVIDED?
    if(argc < 0) {
	fprintf(stderr, "Usage: .....\n");
        exit(EXIT_FAILURE);
    }

//  LIST THE CURRENT DIRECTORY
    if(argc == 1) {
        scan_directory(".");
    }
    else {
//  SCAN EVERY DIRECTORY REQUESTED ON THE COMMAND-LINE
	for(int a=1 ; a<argc ; ++a) {
	    scan_directory(argv[a]);
	}
    }

//  SORT OUR ARRAY OF STRUCTURES BY MODIFICATION TIME
    qsort((void *)files, nfiles, sizeof(FILES),
            reverse ? compare_mtime_descending : compare_mtime_ascending);

    list_all_files();

    return 0;
}
